@@ -12,12 +12,11 @@
 #include <string.h>
 #include "page.h"
 
-
 enum {
-    IMAGE_CHANGED,
+    PIXELS_CHANGED,
     SIZE_CHANGED,
     SCAN_LINE_CHANGED,
-    ORIENTATION_CHANGED,
+    SCAN_DIRECTION_CHANGED,
     CROP_CHANGED,
     LAST_SIGNAL
 };
@@ -29,11 +28,21 @@ struct PagePrivate
     gint dpi;
 
     /* Number of rows in this page or -1 if currently unknown */
-    gint rows;
+    gint expected_rows;
+
+    /* Bit depth */
+    gint depth;
+
+    /* Color profile */
+    gchar *color_profile;
 
     /* Scanned image data */
-    GdkPixbuf *image;
-  
+    gint width, n_rows, rowstride, n_channels;
+    guchar *pixels;
+
+    /* Page is getting data */
+    gboolean scanning;
+
     /* TRUE if have some page data */
     gboolean has_data;
 
@@ -41,8 +50,8 @@ struct PagePrivate
     gint scan_line;
 
     /* Rotation of scanned data */
-    Orientation orientation;
-    
+    ScanDirection scan_direction;
+
     /* Crop */
     gboolean has_crop;
     gchar *crop_name;
@@ -53,45 +62,54 @@ G_DEFINE_TYPE (Page, page, G_TYPE_OBJECT);
 
 
 Page *
-page_new ()
+page_new (gint width, gint height, gint dpi, ScanDirection scan_direction)
 {
-    return g_object_new (PAGE_TYPE, NULL);
+    Page *page;
+  
+    page = g_object_new (PAGE_TYPE, NULL);
+    if (scan_direction == TOP_TO_BOTTOM || scan_direction == BOTTOM_TO_TOP) {
+        page->priv->width = width;
+        page->priv->n_rows = height;
+    }
+    else {
+        page->priv->width = height;
+        page->priv->n_rows = width;
+    }
+    page->priv->dpi = dpi;
+    page->priv->scan_direction = scan_direction;
+
+    return page;
 }
 
 
 void
-page_set_scan_area (Page *page, gint width, gint rows, gint dpi)
+page_set_page_info (Page *page, ScanPageInfo *info)
 {
-    gint height;
-
     g_return_if_fail (page != NULL);
 
-    /* Variable height, try 50% of the width for now */
-    if (rows < 0)
-        height = width / 2;
-    else
-        height = rows;
-
-    /* Rotate page */
-    if (page->priv->orientation == LEFT_TO_RIGHT || page->priv->orientation == RIGHT_TO_LEFT) {
-        gint t;
-        t = width;
-        width = height;
-        height = t;
-    }
-
-    page->priv->rows = rows;
-    page->priv->dpi = dpi;
+    page->priv->expected_rows = info->height;
+    page->priv->dpi = info->dpi;
 
     /* Create a white page */
-    /* NOTE: Pixbuf only supports 8 bit RGB images */
-    page->priv->image = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE,
-                                        8,
-                                        width,
-                                        height);
-    gdk_pixbuf_fill (page->priv->image, 0xFFFFFFFF);
+    page->priv->width = info->width;
+    page->priv->n_rows = info->height;
+    /* Variable height, try 50% of the width for now */
+    if (page->priv->n_rows < 0)
+        page->priv->n_rows = page->priv->width / 2;
+    page->priv->depth = info->depth;
+    page->priv->n_channels = info->n_channels;
+    page->priv->rowstride = (page->priv->width * page->priv->depth * page->priv->n_channels + 7) / 8;
+    page->priv->pixels = g_realloc (page->priv->pixels, page->priv->n_rows * page->priv->rowstride);
+    g_return_if_fail (page->priv->pixels != NULL);
+
+    /* Fill with white */
+    if (page->priv->depth == 1)
+        memset (page->priv->pixels, 0x00, page->priv->n_rows * page->priv->rowstride);
+    else
+        memset (page->priv->pixels, 0xFF, page->priv->n_rows * page->priv->rowstride);
+
     g_signal_emit (page, signals[SIZE_CHANGED], 0);
-    g_signal_emit (page, signals[IMAGE_CHANGED], 0);
+    g_signal_emit (page, signals[PIXELS_CHANGED], 0);
 }
 
 
@@ -100,172 +118,66 @@ page_start (Page *page)
 {
     g_return_if_fail (page != NULL);
 
-    page->priv->scan_line = 0;
+    page->priv->scanning = TRUE;
     g_signal_emit (page, signals[SCAN_LINE_CHANGED], 0);
 }
 
 
-static gint
-get_sample (guchar *data, gint depth, gint index)
+gboolean
+page_is_scanning (Page *page)
 {
-    gint i, offset, value, n_bits;
-
-    /* Optimise if using 8 bit samples */
-    if (depth == 8)
-        return data[index];
-
-    /* Bit offset for this sample */
-    offset = depth * index;
-
-    /* Get the remaining bits in the octet this sample starts in */
-    i = offset / 8;
-    n_bits = 8 - offset % 8;
-    value = data[i] & (0xFF >> (8 - n_bits));
-    
-    /* Add additional octets until get enough bits */
-    while (n_bits < depth) {
-        value = value << 8 | data[i++];
-        n_bits += 8;
-    }
-
-    /* Trim remaining bits off */
-    if (n_bits > depth)
-        value >>= n_bits - depth;
-
-    return value;
+    g_return_val_if_fail (page != NULL, FALSE);
+  
+    return page->priv->scanning;
 }
 
 
-gboolean page_has_data (Page *page)
+gboolean
+page_has_data (Page *page)
 {
     g_return_val_if_fail (page != NULL, FALSE);
     return page->priv->has_data;
 }
 
 
-gint page_get_scan_line (Page *page)
+gboolean
+page_is_color (Page *page)
+{
+    g_return_val_if_fail (page != NULL, FALSE);
+    return page->priv->n_channels > 1;
+}
+
+
+gint
+page_get_scan_line (Page *page)
 {
     g_return_val_if_fail (page != NULL, -1);
     return page->priv->scan_line;
 }
 
 
-static void
-set_pixel (ScanLine *line, gint n, gint x, guchar *pixel)
-{
-    gint sample;
-    guchar *data;
-  
-    data = line->data + line->data_length * n;
-
-    switch (line->format) {
-    case LINE_RGB:
-        pixel[0] = get_sample (data, line->depth, x*3) * 0xFF / ((1 << line->depth) - 1);
-        pixel[1] = get_sample (data, line->depth, x*3+1) * 0xFF / ((1 << line->depth) - 1);
-        pixel[2] = get_sample (data, line->depth, x*3+2) * 0xFF / ((1 << line->depth) - 1);
-        break;
-    case LINE_GRAY:
-        /* Bitmap, 0 = white, 1 = black */
-        sample = get_sample (data, line->depth, x) * 0xFF / ((1 << line->depth) - 1);
-        if (line->depth == 1)
-            sample = sample ? 0x00 : 0xFF;
-
-        pixel[0] = pixel[1] = pixel[2] = sample;
-        break;
-    case LINE_RED:
-        pixel[0] = get_sample (data, line->depth, x) * 0xFF / ((1 << line->depth) - 1);
-        break;
-    case LINE_GREEN:
-        pixel[1] = get_sample (data, line->depth, x) * 0xFF / ((1 << line->depth) - 1);
-        break;
-    case LINE_BLUE:
-        pixel[2] = get_sample (data, line->depth, x) * 0xFF / ((1 << line->depth) - 1);
-        break;
-    }
-}
-
-
 static void
 parse_line (Page *page, ScanLine *line, gint n, gboolean *size_changed)
 {
-    guchar *pixels;
     gint line_number;
-    gint i, x = 0, y = 0, x_step = 0, y_step = 0;
-    gint rowstride, n_channels;
 
     line_number = line->number + n;
 
     /* Extend image if necessary */
     while (line_number >= page_get_scan_height (page)) {
-        GdkPixbuf *image;
-        gint height, width, new_width, new_height;
+        gint rows;
 
         /* Extend image */
-        new_width = width = gdk_pixbuf_get_width (page->priv->image);
-        new_height = height = gdk_pixbuf_get_height (page->priv->image);
-        if (page->priv->orientation == TOP_TO_BOTTOM || page->priv->orientation == BOTTOM_TO_TOP) {
-            new_height = height + width / 2;
-            g_debug("Extending image height from %d pixels to %d pixels", height, new_height);
-        }
-        else {
-            new_width = width + height / 2;
-            g_debug("Extending image width from %d pixels to %d pixels", width, new_width);
-        }
-        image = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE,
-                                8, new_width, new_height);
-
-        /* Copy old data */
-        gdk_pixbuf_fill (image, 0xFFFFFFFF);
-        if (page->priv->orientation == TOP_TO_BOTTOM || page->priv->orientation == LEFT_TO_RIGHT)
-            gdk_pixbuf_copy_area (page->priv->image, 0, 0, width, height,
-                                  image, 0, 0);
-        else
-            gdk_pixbuf_copy_area (page->priv->image, 0, 0, width, height,
-                                  image, new_width - width, new_height - height);
-
-        g_object_unref (page->priv->image);
-        page->priv->image = image;
+        rows = page->priv->n_rows;
+        page->priv->n_rows = rows + page->priv->width / 2;
+        g_debug("Extending image from %d lines to %d lines", rows, page->priv->n_rows);
+        page->priv->pixels = g_realloc (page->priv->pixels, page->priv->n_rows * page->priv->rowstride);
 
         *size_changed = TRUE;
     }
-  
-    switch (page->priv->orientation) {
-    case TOP_TO_BOTTOM:
-        x = 0;
-        y = line_number;
-        x_step = 1;
-        y_step = 0;
-        break;
-    case BOTTOM_TO_TOP:
-        x = page_get_width (page) - 1;
-        y = page_get_height (page) - line_number - 1;
-        x_step = -1;
-        y_step = 0;
-        break;
-    case LEFT_TO_RIGHT:
-        x = line_number;
-        y = page_get_height (page) - 1;
-        x_step = 0;
-        y_step = -1;
-        break;
-    case RIGHT_TO_LEFT:
-        x = page_get_width (page) - line_number - 1;
-        y = 0;
-        x_step = 0;
-        y_step = 1;
-        break;
-    }
-    pixels = gdk_pixbuf_get_pixels (page->priv->image);
-    rowstride = gdk_pixbuf_get_rowstride (page->priv->image);
-    n_channels = gdk_pixbuf_get_n_channels (page->priv->image);
-    for (i = 0; i < line->width; i++) {
-        guchar *pixel;
 
-        pixel = pixels + y * rowstride + x * n_channels;
-        set_pixel (line, n, i, pixel);
-        x += x_step;
-        y += y_step;
-    }
+    /* Copy in new row */
+    memcpy (page->priv->pixels + line_number * page->priv->rowstride, line->data + n * line->data_length, line->data_length);
 
     page->priv->scan_line = line_number;
 }
@@ -287,7 +199,7 @@ page_parse_scan_line (Page *page, ScanLine *line)
     if (size_changed)
         g_signal_emit (page, signals[SIZE_CHANGED], 0);
     g_signal_emit (page, signals[SCAN_LINE_CHANGED], 0);
-    g_signal_emit (page, signals[IMAGE_CHANGED], 0);
+    g_signal_emit (page, signals[PIXELS_CHANGED], 0);
 }
 
 
@@ -299,38 +211,18 @@ page_finish (Page *page)
     g_return_if_fail (page != NULL);
 
     /* Trim page */
-    if (page->priv->rows < 0 &&
-        page->priv->scan_line != gdk_pixbuf_get_height (page->priv->image)) {
-        GdkPixbuf *image;
-        gint width, height, new_width, new_height;
+    if (page->priv->expected_rows < 0 &&
+        page->priv->scan_line != page_get_scan_height (page)) {
+        gint rows;
 
-        new_width = width = gdk_pixbuf_get_width (page->priv->image);
-        new_height = height = gdk_pixbuf_get_height (page->priv->image);
-        if (page->priv->orientation == TOP_TO_BOTTOM || page->priv->orientation == BOTTOM_TO_TOP) {
-            new_height = page->priv->scan_line;
-            g_debug("Trimming image height from %d pixels to %d pixels", height, new_height);
-        }
-        else {
-            new_width = page->priv->scan_line;
-            g_debug("Trimming image width from %d pixels to %d pixels", width, new_width);          
-        }
-        image = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE,
-                                8,
-                                new_width, new_height);
+        rows = page->priv->n_rows;
+        page->priv->n_rows = page->priv->scan_line;
+        page->priv->pixels = g_realloc (page->priv->pixels, page->priv->n_rows * page->priv->rowstride);
+        g_debug("Trimming page from %d lines to %d lines", rows, page->priv->n_rows);
 
-        /* Copy old data */
-        if (page->priv->orientation == TOP_TO_BOTTOM || page->priv->orientation == LEFT_TO_RIGHT)
-            gdk_pixbuf_copy_area (page->priv->image, 0, 0, width, height,
-                                  image, 0, 0);
-        else
-            gdk_pixbuf_copy_area (page->priv->image, width - new_width, height - new_height, width, height,
-                                  image, 0, 0);
-
-        g_object_unref (page->priv->image);
-        page->priv->image = image;
         size_changed = TRUE;
     }
-    page->priv->scan_line = -1;
+    page->priv->scanning = FALSE;
 
     if (size_changed)
         g_signal_emit (page, signals[SIZE_CHANGED], 0);
@@ -338,45 +230,36 @@ page_finish (Page *page)
 }
 
 
-Orientation
-page_get_orientation (Page *page)
+ScanDirection
+page_get_scan_direction (Page *page)
 {
     g_return_val_if_fail (page != NULL, TOP_TO_BOTTOM);
 
-    return page->priv->orientation;
+    return page->priv->scan_direction;
 }
 
 
-void
-page_set_orientation (Page *page, Orientation orientation)
+static void
+page_set_scan_direction (Page *page, ScanDirection scan_direction)
 {
     gint left_steps, t;
-    GdkPixbuf *image;
     gboolean size_changed = FALSE;
+    gint width, height;
 
     g_return_if_fail (page != NULL);
 
-    if (page->priv->orientation == orientation)
+    if (page->priv->scan_direction == scan_direction)
         return;
 
     /* Work out how many times it has been rotated to the left */
-    left_steps = orientation - page->priv->orientation;
+    left_steps = scan_direction - page->priv->scan_direction;
     if (left_steps < 0)
         left_steps += 4;
-  
-    /* Rotate image */
-    if (left_steps == 1)
-        image = gdk_pixbuf_rotate_simple (page->priv->image, GDK_PIXBUF_ROTATE_COUNTERCLOCKWISE);
-    else if (left_steps == 2)
-        image = gdk_pixbuf_rotate_simple (page->priv->image, GDK_PIXBUF_ROTATE_UPSIDEDOWN);
-    else if (left_steps == 3)
-        image = gdk_pixbuf_rotate_simple (page->priv->image, GDK_PIXBUF_ROTATE_CLOCKWISE);
-    else
-        image = gdk_pixbuf_rotate_simple (page->priv->image, GDK_PIXBUF_ROTATE_NONE);
-    gdk_pixbuf_unref (page->priv->image);
-    page->priv->image = image;
     if (left_steps != 2)
         size_changed = TRUE;
+  
+    width = page_get_width (page);
+    height = page_get_height (page);
 
     /* Rotate crop */
     if (page->priv->has_crop) {
@@ -385,21 +268,21 @@ page_set_orientation (Page *page, Orientation orientation)
         case 1:
             t = page->priv->crop_x;
             page->priv->crop_x = page->priv->crop_y;
-            page->priv->crop_y = page_get_width (page) - (t + page->priv->crop_width);
+            page->priv->crop_y = width - (t + page->priv->crop_width);
             t = page->priv->crop_width;
             page->priv->crop_width = page->priv->crop_height;
             page->priv->crop_height = t;
             break;
         /* 180 degrees */
         case 2:
-            page->priv->crop_x = page_get_width (page) - (page->priv->crop_x + page->priv->crop_width);
-            page->priv->crop_y = page_get_width (page) - (page->priv->crop_y + page->priv->crop_height);
+            page->priv->crop_x = width - (page->priv->crop_x + page->priv->crop_width);
+            page->priv->crop_y = width - (page->priv->crop_y + page->priv->crop_height);
             break;
         /* 90 degrees clockwise */
         case 3:
             t = page->priv->crop_y;
             page->priv->crop_y = page->priv->crop_x;
-            page->priv->crop_x = page_get_height (page) - (t + page->priv->crop_height);
+            page->priv->crop_x = height - (t + page->priv->crop_height);
             t = page->priv->crop_width;
             page->priv->crop_width = page->priv->crop_height;
             page->priv->crop_height = t;
@@ -407,42 +290,42 @@ page_set_orientation (Page *page, Orientation orientation)
         }
     }
 
-    page->priv->orientation = orientation;
+    page->priv->scan_direction = scan_direction;
     if (size_changed)
-        g_signal_emit (page, signals[SIZE_CHANGED], 0);  
-    g_signal_emit (page, signals[IMAGE_CHANGED], 0);
-    g_signal_emit (page, signals[ORIENTATION_CHANGED], 0);
-    g_signal_emit (page, signals[CROP_CHANGED], 0);
+        g_signal_emit (page, signals[SIZE_CHANGED], 0);
+    g_signal_emit (page, signals[SCAN_DIRECTION_CHANGED], 0);
+    if (page->priv->has_crop)
+        g_signal_emit (page, signals[CROP_CHANGED], 0);
 }
 
 
 void
 page_rotate_left (Page *page)
 {
-    Orientation orientation;
+    ScanDirection scan_direction;
 
     g_return_if_fail (page != NULL);
 
-    orientation = page_get_orientation (page);
-    if (orientation == RIGHT_TO_LEFT)
-        orientation = TOP_TO_BOTTOM;
+    scan_direction = page_get_scan_direction (page);
+    if (scan_direction == RIGHT_TO_LEFT)
+        scan_direction = TOP_TO_BOTTOM;
     else
-        orientation++;
-    page_set_orientation (page, orientation);
+        scan_direction++;
+    page_set_scan_direction (page, scan_direction);
 }
 
 
 void
 page_rotate_right (Page *page)
 {
-    Orientation orientation;
+    ScanDirection scan_direction;
 
-    orientation = page_get_orientation (page);
-    if (orientation == TOP_TO_BOTTOM)
-        orientation = RIGHT_TO_LEFT;
+    scan_direction = page_get_scan_direction (page);
+    if (scan_direction == TOP_TO_BOTTOM)
+        scan_direction = RIGHT_TO_LEFT;
     else
-        orientation--;
-    page_set_orientation (page, orientation);
+        scan_direction--;
+    page_set_scan_direction (page, scan_direction);
 }
 
 
@@ -466,7 +349,11 @@ gint
 page_get_width (Page *page)
 {
     g_return_val_if_fail (page != NULL, 0);
-    return gdk_pixbuf_get_width (page->priv->image);
+
+    if (page->priv->scan_direction == TOP_TO_BOTTOM || page->priv->scan_direction == BOTTOM_TO_TOP)
+        return page->priv->width;
+    else
+        return page->priv->n_rows;
 }
 
 
@@ -474,7 +361,33 @@ gint
 page_get_height (Page *page)
 {
     g_return_val_if_fail (page != NULL, 0);
-    return gdk_pixbuf_get_height (page->priv->image);
+
+    if (page->priv->scan_direction == TOP_TO_BOTTOM || page->priv->scan_direction == BOTTOM_TO_TOP)
+        return page->priv->n_rows;
+    else
+        return page->priv->width;
+}
+
+
+gint
+page_get_depth (Page *page)
+{
+    g_return_val_if_fail (page != NULL, 0);
+    return page->priv->depth;
+}
+
+
+gint page_get_n_channels (Page *page)
+{
+    g_return_val_if_fail (page != NULL, 0);
+    return page->priv->n_channels;
+}
+
+
+gint page_get_rowstride (Page *page)
+{
+    g_return_val_if_fail (page != NULL, 0);
+    return page->priv->rowstride;
 }
 
 
@@ -483,10 +396,7 @@ page_get_scan_width (Page *page)
 {
     g_return_val_if_fail (page != NULL, 0);
 
-    if (page->priv->orientation == TOP_TO_BOTTOM || page->priv->orientation == BOTTOM_TO_TOP)
-        return gdk_pixbuf_get_width (page->priv->image);
-    else
-        return gdk_pixbuf_get_height (page->priv->image);
+    return page->priv->width;
 }
 
 
@@ -495,11 +405,20 @@ page_get_scan_height (Page *page)
 {
     g_return_val_if_fail (page != NULL, 0);
 
-    if (page->priv->orientation == TOP_TO_BOTTOM || page->priv->orientation == BOTTOM_TO_TOP)
-        return gdk_pixbuf_get_height (page->priv->image);
-    else
-        return gdk_pixbuf_get_width (page->priv->image);  
+    return page->priv->n_rows;
+}
 
+
+void page_set_color_profile (Page *page, const gchar *color_profile)
+{
+     g_free (page->priv->color_profile);
+     page->priv->color_profile = g_strdup (color_profile);
+}
+
+
+const gchar *page_get_color_profile (Page *page)
+{
+     return page->priv->color_profile;
 }
 
 
@@ -703,60 +622,136 @@ page_get_named_crop (Page *page)
 }
 
 
-GdkPixbuf *
-page_get_image (Page *page)
+const guchar *
+page_get_pixels (Page *page)
 {
     g_return_val_if_fail (page != NULL, NULL);
-    return gdk_pixbuf_copy (page->priv->image);
+    return page->priv->pixels;
+}
+
+
+// FIXME: Copied from page-view, should be shared code
+static guchar
+get_sample (const guchar *line, gint x, gint depth, gint n_channels, gint channel)
+{
+    // FIXME
+    return 0xFF;
+}
+
+
+// FIXME: Copied from page-view, should be shared code
+static void
+get_pixel (Page *page, gint x, gint y, guchar *pixel)
+{
+    gint t, depth, n_channels;
+    const guchar *p, *line;
+
+    switch (page_get_scan_direction (page))
+    {
+    case TOP_TO_BOTTOM:
+        break;
+    case BOTTOM_TO_TOP:
+        x = page_get_scan_width (page) - x;
+        y = page_get_scan_height (page) - y;
+        break;
+    case LEFT_TO_RIGHT:
+        t = x;
+        x = page_get_scan_width (page) - y;
+        y = t;
+        break;
+    case RIGHT_TO_LEFT:
+        t = x;
+        x = y;
+        y = page_get_scan_height (page) - t;
+        break;
+    }
+
+    depth = page_get_depth (page);
+    n_channels = page_get_n_channels (page);
+    line = page_get_pixels (page) + page_get_rowstride (page) * y;
+
+    /* Optimise for 8 bit images */
+    if (depth == 8 && n_channels == 3) {
+        p = line + x * n_channels;
+        pixel[0] = p[0];
+        pixel[1] = p[1];
+        pixel[2] = p[2];
+        return;
+    }
+    else if (depth == 8 && n_channels == 1) {
+        p = line + x;
+        pixel[0] = pixel[1] = pixel[2] = p[0];
+        return;
+    }
+
+    /* Optimise for bitmaps */
+    else if (depth == 1 && n_channels == 1) {
+        p = line + (x / 8);
+        pixel[0] = pixel[1] = pixel[2] = p[0] & (0x80 >> (x % 8)) ? 0x00 : 0xFF;
+        return;
+    }
+
+    /* Optimise for 2 bit images */
+    else if (depth == 2 && n_channels == 1) {
+        gint sample;
+        gint block_shift[4] = { 6, 4, 2, 0 };
+
+        p = line + (x / 4);
+        sample = (p[0] >> block_shift[x % 4]) & 0x3;
+        sample = sample * 255 / 3;
+
+        pixel[0] = pixel[1] = pixel[2] = sample;
+        return;
+    }
+
+    /* Use slow method */
+    pixel[0] = get_sample (line, x, depth, n_channels, 0);
+    pixel[0] = get_sample (line, x, depth, n_channels, 1);
+    pixel[0] = get_sample (line, x, depth, n_channels, 2);
 }
 
 
 GdkPixbuf *
-page_get_cropped_image (Page *page)
+page_get_image (Page *page, gboolean apply_crop)
 {
-    GdkPixbuf *image, *cropped_image, *i;
-    gint x, y, w, h, pw, ph;
+    GdkPixbuf *image;
+    gint x, y, l, r, t, b;
 
-    g_return_val_if_fail (page != NULL, NULL);
-    
-    image = page_get_image (page);
-    
-    if (!page->priv->has_crop)
-        return image;
-    
-    x = page->priv->crop_x;
-    y = page->priv->crop_y;
-    w = page->priv->crop_width;
-    h = page->priv->crop_height;
-    pw = gdk_pixbuf_get_width (image);
-    ph = gdk_pixbuf_get_height (image);
-    
-    /* Trim crop */
-    if (x + w >= pw)
-        w = pw - x;
-    if (y + h >= ph)
-        h = ph - y;
-    
-    cropped_image = gdk_pixbuf_new_subpixbuf (image, x, y, w, h);
-    g_object_unref (image);
-    
-    i = gdk_pixbuf_copy (cropped_image);
-    g_object_unref (cropped_image);
+    if (apply_crop && page->priv->has_crop) {
+        l = page->priv->crop_x;
+        r = l + page->priv->crop_width;
+        t = page->priv->crop_y;
+        b = l + page->priv->crop_height;
+      
+        if (l < 0)
+            l = 0;
+        if (r > page_get_width (page))
+            r = page_get_width (page);
+        if (t < 0)
+            t = 0;
+        if (b > page_get_height (page))
+            b = page_get_height (page);
+    }
+    else {
+        l = 0;
+        r = page_get_width (page);
+        t = 0;
+        b = page_get_height (page);
+    }
 
-    return i;
-}
+    image = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, r - l, b - t);
 
+    for (y = t; y < b; y++) {
+        guchar *line = gdk_pixbuf_get_pixels (image) + gdk_pixbuf_get_rowstride (image) * (y - t);
+        for (x = l; x < r; x++) {
+            guchar *pixel;
 
-static GFileOutputStream *
-open_file (const gchar *uri, GError **error)
-{
-    GFile *file;
-    GFileOutputStream *stream; 
+            pixel = line + (x - l) * 3;
+            get_pixel (page, x, y, pixel);
+        }
+    }
 
-    file = g_file_new_for_uri (uri);
-    stream = g_file_replace (file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, error);
-    g_object_unref (file);
-    return stream;
+    return image;
 }
 
 
@@ -767,46 +762,78 @@ write_pixbuf_data (const gchar *buf, gsize count, GError **error, GFileOutputStr
 }
 
 
-gboolean
-page_save_jpeg (Page *page, const gchar *uri, GError **error)
+static gchar *
+get_icc_data_encoded (const gchar *icc_profile_filename)
 {
-    GdkPixbuf *image;
-    GFileOutputStream *stream;
-    gboolean result;
+    gchar *contents = NULL;
+    gchar *contents_encode = NULL;
+    gsize length;
+    gboolean ret;
+    GError *error = NULL;
+
+    /* Get binary data */
+    ret = g_file_get_contents (icc_profile_filename, &contents, &length, &error);
+    if (!ret) {
+        g_warning ("failed to get icc profile data: %s", error->message);
+        g_error_free (error);
+    }
+    else {
+        /* Encode into base64 */
+        contents_encode = g_base64_encode ((const guchar *) contents, length);
+    }
   
-    stream = open_file (uri, error);
-    if (!stream)
-        return FALSE;
-
-    image = page_get_cropped_image (page);
-    result = gdk_pixbuf_save_to_callback (image,
-                                          (GdkPixbufSaveFunc) write_pixbuf_data, stream,
-                                          "jpeg", error,
-                                          "quality", "90",
-                                          NULL);
-    g_object_unref (image);
-    g_object_unref (stream);
-
-    return result;
+    g_free (contents);
+    return contents_encode;
 }
 
 
 gboolean
-page_save_png (Page *page, const gchar *uri, GError **error)
+page_save (Page *page, const gchar *type, GFile *file, GError **error)
 {
-    GdkPixbuf *image;
     GFileOutputStream *stream;
-    gboolean result;
-  
-    stream = open_file (uri, error);
+    GdkPixbuf *image;
+    gboolean result = FALSE;
+    gchar *icc_profile_data = NULL;
+
+    stream = g_file_replace (file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, error);
     if (!stream)
         return FALSE;
 
-    image = page_get_cropped_image (page);
-    result = gdk_pixbuf_save_to_callback (image,
-                                          (GdkPixbufSaveFunc) write_pixbuf_data, stream,
-                                          "png", error,
-                                          NULL);
+    image = page_get_image (page, TRUE);
+
+    if (page->priv->color_profile != NULL)
+        icc_profile_data = get_icc_data_encoded (page->priv->color_profile);
+
+    if (strcmp (type, "jpeg") == 0) {
+        /* ICC profile is awaiting review in gtk2+ bugzilla */
+        gchar *keys[] = { "quality", /* "icc-profile", */ NULL };
+        gchar *values[] = { "90", /* icc_profile_data, */ NULL };
+        result = gdk_pixbuf_save_to_callbackv (image,
+                                               (GdkPixbufSaveFunc) write_pixbuf_data, stream,
+                                               "jpeg", keys, values, error);
+    }
+    else if (strcmp (type, "png") == 0) {
+        gchar *keys[] = { "icc-profile", NULL };
+        gchar *values[] = { icc_profile_data, NULL };
+        if (icc_profile_data == NULL)
+            keys[0] = NULL;
+        result = gdk_pixbuf_save_to_callbackv (image,
+                                               (GdkPixbufSaveFunc) write_pixbuf_data, stream,
+                                               "png", keys, values, error);
+    }
+    else if (strcmp (type, "tiff") == 0) {
+        gchar *keys[] = { "compression", "icc-profile", NULL };
+        gchar *values[] = { "8" /* Deflate compression */, icc_profile_data, NULL };
+        if (icc_profile_data == NULL)
+            keys[1] = NULL;
+        result = gdk_pixbuf_save_to_callbackv (image,
+                                               (GdkPixbufSaveFunc) write_pixbuf_data, stream,
+                                               "tiff", keys, values, error);
+    }
+    else
+        result = FALSE; // FIXME: Set GError
+
+    g_free (icc_profile_data);
     g_object_unref (image);
     g_object_unref (stream);
 
@@ -818,8 +845,7 @@ static void
 page_finalize (GObject *object)
 {
     Page *page = PAGE (object);
-    g_object_unref (page->priv->image);
-    page->priv->image = NULL;
+    g_free (page->priv->pixels);
     G_OBJECT_CLASS (page_parent_class)->finalize (object);
 }
 
@@ -831,11 +857,11 @@ page_class_init (PageClass *klass)
 
     object_class->finalize = page_finalize;
 
-    signals[IMAGE_CHANGED] =
-        g_signal_new ("image-changed",
+    signals[PIXELS_CHANGED] =
+        g_signal_new ("pixels-changed",
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
-                      G_STRUCT_OFFSET (PageClass, image_changed),
+                      G_STRUCT_OFFSET (PageClass, pixels_changed),
                       NULL, NULL,
                       g_cclosure_marshal_VOID__VOID,
                       G_TYPE_NONE, 0);
@@ -855,11 +881,11 @@ page_class_init (PageClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__VOID,
                       G_TYPE_NONE, 0);
-    signals[ORIENTATION_CHANGED] =
-        g_signal_new ("orientation-changed",
+    signals[SCAN_DIRECTION_CHANGED] =
+        g_signal_new ("scan-direction-changed",
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
-                      G_STRUCT_OFFSET (PageClass, orientation_changed),
+                      G_STRUCT_OFFSET (PageClass, scan_direction_changed),
                       NULL, NULL,
                       g_cclosure_marshal_VOID__VOID,
                       G_TYPE_NONE, 0);
@@ -880,6 +906,5 @@ static void
 page_init (Page *page)
 {
     page->priv = G_TYPE_INSTANCE_GET_PRIVATE (page, PAGE_TYPE, PagePrivate);
-    page->priv->scan_line = -1;
-    page->priv->orientation = TOP_TO_BOTTOM;
+    page->priv->scan_direction = TOP_TO_BOTTOM;
 }
